local timer = {}
timer.clock = Timer()
timer.desc = nil --"Timer has expired" --default message
timer.start_time = 0 --game time timer started (msec)
timer.duration = 0 --length of timer (msec)
timer.is_msec = false --false if resuming old timer on reload using os.time --TODO move to main

timer.params = {} --list of timer settings
timer.pending = 0 --remaining count (msec) for incremental timers

timer.CHAT_TYPES = {sector=true, group=true, hud=true, none=true}
timer.INC_VALS = {yes=1,no=0,on=1,off=0,['true']=1,['false']=0}

--[[ Creates a new timer
* Arguments same as for timer:set()
-> (table) returns the new timer ]]
function timer:new(args)
	--create new timer
	local to_copy = {new=true,set=true,kill=true,msg=true,elapsed=true,remaining=true,complete=true,increment=true,CHAT_TYPES=true,INC_VALS=true,GetNextMilestone=true}
	local new_timer = {}
	new_timer.clock = Timer()
	for k,v in pairs(to_copy) do
		if v then new_timer[k] = self[k] end
	end
	
	if args then --if args specified then start timer
		new_timer:set(args) --start countdown for new timer
		new_timer.desc = new_timer.desc or ((not new_timer.params or not new_timer.params.is_increment) and "Timer has expired") or nil
	end
	
	return new_timer
end

--[[ Set time for timer and activate it
* (string) [optional] Chat type to ouput to (e.g. sector or group)
* (string) [optional] parameter to set along with value, separated by = or : (e.g. increment=yes). If value not specified then use yes.
* (string) length of the timer: numbers delimited by colons to separate seconds, minutes, hours, days. If only one number given use minutes, except use seconds if incremental
* (string) [optional] Message to show when timer completes. Multiple arguments are concatenated together
-> (number) duration = Returns length of the timer in msec ]]
function timer:set(args)
	--Converts string of form 'key=value' or 'key:value' to table containing key and value
	local function get_param(str)
		if type(str)~='string' then return end
		
		local key,val = str:match('^(.+)[:=](.+)$') --extract key and value from string
		
		if key=='increment' then
			local bool = (val or nil) and self.INC_VALS[val]==1
			return bool~=nil and {key="is_increment",value=bool} or {}
		elseif str=='increment' then return {key="is_increment",value=true} --if value not specified use true
		else return {} end
	end
	
	local temp_params = {} --don't replace existing parameters until all args known valid
	
	--extract parameters from args
	if not args then return end
	if args[1] then
		args[1] = args[1]:lower()
		if self.CHAT_TYPES[ args[1] ] then
			temp_params.output = args[1]
			table.remove(args, 1)
		end
		
		local param = args[1] and get_param(args[1])
		while param and param.key do
			temp_params[param.key] = param.value
			table.remove(args, 1)
			param = args[1] and get_param(args[1])
		end
	end
	
	if temp_params.is_increment==nil then temp_params.is_increment = not not temp_params.output end --enable increment by default if chat output is specified
	
	if args[1] then --length of timer
		self.duration = Choco.util.parseTime(args[1], not not temp_params.output and 'seconds' or 'minutes') --default of seconds if chat output specified, otherwise default minutes
		if self.duration then self.clock:Kill() else return end --kill existing timer
		
		self.start_time = gkmisc.GetGameTime()
		self.params = temp_params
		
		if self.params and self.params.is_increment then
			local pending = self.GetNextMilestone(self.duration) --first milestone
			
			self.pending = pending
			self.clock:SetTimeout(math.max(self.duration-pending, 1), function() return pending>0 and self:increment() or self:complete() end) --start timer to next milestone
			
			if self.params and self.params.output then
				local data = {id=self.start_time,duration=self.duration,desc=self.desc,params=self.params,pending=self.duration}
				ProcessEvent("CHOCO_TIMER_INCREMENT",data)
			end
		else
			self.pending = 0 --non-incremental timer
			self.clock:SetTimeout(self.duration, function() self:complete() end) --start timer
		end
		
		self.is_msec = true
		
		table.remove(args, 1)
		if args[1] then self:msg(args) end --remaining args used for message
		
		return self.duration
	else return end
end

--[[ Stops timer if active
-> (boolean) Returns true if the timer was active and was stopped ]]
function timer:kill()
	if self.clock:IsActive() then
		self.clock:Kill()
		self.start_time = gkmisc.GetGameTime()
		self.duration = 0
		self.is_msec = true
		return true
	else return false end
end

--[[ Set message to display when timer completes
* (string) Message to set. Multiple arguments are concatenated together
-> (string) Returns the newly set message if successful ]]
function timer:msg(args)
	if args and args[1] then
		self.desc = table.concat(args, " ") --set message
		return self.desc
	else return end
end

--[[ Get the amount of time since the timer started
-> (number) Returns the time since the timer started in msec if active. Otherwise returns time in msec since timer completed (negative) ]]
function timer:elapsed()
	local msec --number of milliseconds elapsed
	
	if self.start_time > 0 then
		if self.is_msec then
			msec = gkmisc.DiffTime(self.start_time, gkmisc.GetGameTime())
		else msec = os.difftime(os.time(), self.start_time)*1000 end
		
		if self.clock and self.clock:IsActive() then return msec
		else return (self.duration - msec) end --negative
	else return end
end

--[[ Get the amount of time remaining until the timer completes
-> (string) Returns string for time remaining (e.g. XX hours, XX minutes, xx seconds). Omitted if zero ]]
function timer:remaining()
	local msec --number of milliseconds remaining
	
	if self.clock:IsActive() then
		if self.is_msec then
			msec = ( self.duration - gkmisc.DiffTime(self.start_time, gkmisc.GetGameTime()) )
		else msec = self.duration - os.difftime(os.time(), self.start_time)*1000 end --convert os.time() to milliseconds
		
		return Choco.util.msec2str(msec)
	else return end
end

--[[ Called when timer completes. Triggers CHOCO_TIMER_COMPLETED event
-> Returns true so logic statement works out correctly ]]
function timer:complete()
	local data = {id=self.start_time,duration=self.duration,desc=self.desc,params=self.params}
	ProcessEvent("CHOCO_TIMER_COMPLETED",data)
	
	return true
end

--[[ Called when milestone reached for incremental timers. Triggers CHOCO_TIMER_INCREMENT event
-> Returns true so logic statement works out correctly ]]
function timer:increment()
	local data = {id=self.start_time,duration=self.duration,desc=self.desc,params=self.params,pending=self.pending}
	local pending = self.GetNextMilestone(self.pending) --time in msec when next milestone occurs
	
	ProcessEvent("CHOCO_TIMER_INCREMENT",data)
	local remaining = self.duration - gkmisc.DiffTime(self.start_time, gkmisc.GetGameTime()) --recalculate remaining time
	local next_milestone = remaining - pending --compensate for lag
	
	self.clock:SetTimeout(next_milestone, function() return pending>0 and self:increment() or self:complete() end) --set timer until next milestone
	self.pending = pending
		
	return true
end

--[[ Gets remaining time on timer to trigger next milestone
* (number) msec = amount of time remaining on timer in milliseconds
-> (number) What remaining time will be on timer at next milestone in milliseconds ]]
function timer.GetNextMilestone(msec) --TODO make local
	if msec <= 0 then return 0 end --invalid input
	
	local milestones = {3600,900,300,60,15,5,1} --in seconds
	
	for i,sec in ipairs(milestones) do
		local ms = sec*1000 --convert to milliseconds
		
		if ms < msec then
			local diff = msec%ms --remainder to next milestone
			
			if diff==0 then --find difference to next multiple of milestone
				local count = math.floor(msec/ms) - 1 --must be > 0
				local remaining = count*ms
				
				return remaining
			else return msec-diff end
		end
	end
	
	return 0 --no more milestones before timer expires
end

return timer